#!/usr/bin/env python
# encoding: utf-8

"""
funcutils provided some useful functions.
"""

import inspect


def is_magic_method(method):
    """
    Check a method if is magic method

    :param method: method of a class
    """
    if not inspect.ismethod(method):
        return False
    func_name = method.im_func.func_name
    if func_name.startswith("__") and func_name.endswith("__"):
        return True
    return False


def set_default_attr(obj, name, value):
    """
    Set attribute to obj if attribute is not exists

    :param obj: object
    :param name: attribure name
    :param value: attribure value
    :return: if is attribute not exists
    """
    if not hasattr(obj, name):
        setattr(obj, name, value)
    return getattr(obj, name, value)


def iter_attrs(
    obj, include_fields=(), exclude_fields=(),
    public_only=True, exclude_methods=True,
):
    """
    iter attributes of obj

    :param obj: object
    :param include_fields: include all this fields if exists
    :param exclude_fields: exclude this fields
    :param public_only: choice public attr only
    :param exclude_methods: exclude methods
    """
    obj_type = type(obj)
    obj_type_name = obj_type.__name__
    for attr, val in inspect.getmembers(obj):
        if attr not in include_fields:
            if (
                attr in exclude_fields
                or attr.startswith("__")
                or attr.startswith("_%s__" % obj_type_name)
                or (public_only and attr.startswith("_"))
                or (exclude_methods and inspect.ismethod(val))
            ):
                continue
        yield attr, val



#!/usr/bin/env python
# encoding: utf-8

from contextlib import contextmanager


class StateNotAllowError(Exception):
    pass


class EndOfSequenceStateError(Exception):
    pass


class SequenceStateFinishedError(Exception):
    pass


class FreeState(object):
    def __init__(self, init_state):
        self.last_state = None
        self.state = init_state
        self.exception = None

    def check_is_in_states(self, states):
        if self.state not in states:
            raise StateNotAllowError("%s not allow" % self.state)

    def state_to(self, to_state):
        self.last_state = self.state
        self.state = to_state

    def on_success(self, allow_states, success_state):
        self.check_is_in_states(allow_states)
        self.state_to(success_state)
        self.exception = None

    def on_fail(self, allow_states, fail_state, exception):
        self.check_is_in_states(allow_states)
        self.state_to(fail_state)
        self.exception = exception

    @contextmanager
    def processing(self, allow_states, success_state, fail_state):
        if not isinstance(allow_states, (tuple, list, set)):
            allow_states = (allow_states,)

        self.check_is_in_states(allow_states)
        try:
            yield
        except Exception as err:
            self.on_fail(allow_states, fail_state, err)
            raise
        else:
            self.on_success(allow_states, success_state)


class SequenceState(FreeState):
    __SequenceStates__ = ()
    __FailState__ = "error"

    def __init__(self, fail_state=NotImplemented):
        if fail_state is not NotImplemented:
            self.__FailState__ = fail_state
        self.index = 0
        self.base = super(SequenceState, self)
        self.base.__init__(init_state=self.__SequenceStates__[self.index])

    @property
    def next_state(self):
        index = self.index + 1
        if index >= len(self.__SequenceStates__):
            raise EndOfSequenceStateError()
        return self.__SequenceStates__[index]

    @property
    def is_finished(self):
        if not self.index < len(self.__SequenceStates__):
            return True
        if self.state is self.__FailState__:
            return True
        return False

    def on_success(self, allow_states, success_state):
        self.base.on_success(allow_states, success_state)
        self.index += 1

    def processing(self, allow_states=None):
        if self.is_finished:
            raise SequenceStateFinishedError()
        if allow_states is None:
            allow_states = self.state

        return self.base.processing(
            allow_states, self.next_state, self.__FailState__,
        )


class QSequenceState(SequenceState):
    def __init__(self, states, fail_state=NotImplemented):
        self.__SequenceStates__ = states
        super(QSequenceState, self).__init__(fail_state=fail_state)


def SequenceStateBuilder(sequence_states, fail_state=NotImplemented):
    class NewSequenceState(SequenceState):
        __SequenceStates__ = sequence_states
        if fail_state is not NotImplemented:
            __FailState__ = fail_state

    return NewSequenceState



#!/usr/bin/env python
# encoding: utf-8

import os
from os import path as os_path
import codecs
import tempfile
import shutil
from contextlib import contextmanager
import exceptions
from functools import wraps
import hashlib
from datetime import datetime
import re

from ycyc.base import contextutils


def oserror_format(func):
    """
    Decorator that catch WindowsError and reraise it as OSError
    """
    @wraps(func)
    def foo(*args, **kwargs):
        with contextutils.catch(
            getattr(exceptions, "WindowsError", OSError),
            reraise=OSError,
        ):
            return func(*args, **kwargs)
    return foo


@oserror_format
def make_sure_dir_exists(path):
    """
    Create the directories when the path is not exists.

    :param path: path string
    """
    if not os_path.exists(path):
        os.makedirs(path)


@oserror_format
def remove_dir(path, recursion=False):
    """
    Remove the directory.

    :param path: path string
    :param recursion: if delete sub directories and files
    """
    if recursion:
        shutil.rmtree(path)
    else:
        os.removedirs(path)


@oserror_format
def make_sure_dir_empty(path):
    """
    Make sure the directory is empty.

    :param path: path string
    """
    if not os_path.exists(path):
        make_sure_dir_exists(path)
    else:
        for i in os.listdir(path):
            i_path = os_path.join(path, i)
            if os_path.isdir(i_path):
                remove_dir(i_path)
            else:
                os.remove(i_path)


@oserror_format
def make_sure_not_exists(path):
    """
    Make sure the path is not exists.

    :param path: path string
    """
    if not os_path.exists(path):
        return
    if os_path.isdir(path):
        remove_dir(path, recursion=True)
    else:
        os.remove(path)


def safe_open_for_write(fn, encoding="utf-8"):
    """
    Auto create the directories and open a file for write(w+).

    :param fn: file path string
    :param encoding: file encoding
    """
    make_sure_dir_exists(os_path.dirname(fn))
    return codecs.open(fn, "w+", encoding)


def touch_file(path):
    """
    As same as *nix command `touch`

    :param path: path string
    """
    with safe_open_for_write(path):
        pass


def safe_open_for_update(fn, encoding="utf-8"):
    """
    Make sure open a file for update(r+).

    :param fn: file path string
    :param encoding: file encoding
    """
    if not os_path.exists(fn):
        return safe_open_for_write(fn, encoding)
    return codecs.open(fn, "r+", encoding)


def safe_open_for_read(fn, encoding="utf-8"):
    """
    Make sure open a file for read(r).

    :param fn: file path string
    :param encoding: file encoding
    """
    if os_path.exists(fn):
        return codecs.open(fn, "r", encoding)
    return tempfile.TemporaryFile("r")


def choice_one_if_exists(paths, default=None):
    """
    Choice the first path in paths which is exists.
    You can use this to choice application configuration files.
    Example:
        there are some configuration files:
            /etc/conf/lyc.conf -> for production
            ~/conf/lyc.debug.conf -> for developing
            ./conf/demo.conf -> for default configuration
        so you can put you real configuration files to this places
        to identify your real environment for this application.

    >>> real_conf = choice_one_if_exists([
    ...     "/etc/conf/lyc.conf",
    ...     "~/conf/lyc.conf",
    ...     "./conf/demo.conf",
    ... ])

    :param fn: file path string
    :param encoding: file encoding
    """
    isexists = os_path.exists
    for p in paths:
        if isexists(p):
            return p
    return default


@contextmanager
@oserror_format
def cd(path):
    """
    Switch to a directory to excute.

    :param path: directory string
    """
    current_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        with contextutils.catch():
            os.chdir(current_dir)


def bytes_from(path):
    """
    Return the bytes from a file.

    :param path: file path string
    """
    with open(path, "rb") as fp:
        return fp.read()


def text_from(path, encoding="utf-8"):
    """
    Return the text from a file.

    :param path: file path string
    :param encoding: file encoding
    """
    with codecs.open(path, "r", encoding) as fp:
        return fp.read()


def hash_of(path, hash_name):
    """
    Return the hash of a file

    :param path: file path string
    :param hash_name: hash algorithm name(md5/sha1/sha256/sha512)
    """
    hash = hashlib.new(hash_name, bytes_from(path))
    return hash.hexdigest()


def sha1_of(path):
    """
    Return the sha1 hash of a file

    :param path: file path string
    """
    return hash_of(path, "sha1")


def md5_of(path):
    """
    Return the md5 hash of a file

    :param path: file path string
    """
    return hash_of(path, "md5")


def available_file_name(name, replaces="_"):
    """
    Return a available file name that replace some forbidden chars

    :param name: file name
    :param replaces: replaces char
    """
    return re.sub(
        r"[\<\>/\\\|:\"\*\?]",
        replaces,
        name,
    )


def mk_not_existed_path(path):
    """
    Return a new file name based on path, and make sure the new name
    is not existed.

    :param path: file name
    """
    name_without_ext, ext = os_path.splitext(path)

    n = 1
    while os_path.exists(path):
        n += 1
        path = name_without_ext + "(%d)" % n + ext

    return path


class PathInfo(object):
    def __init__(self, path):
        self.path = os_path.realpath(path)

    def __str__(self):
        return "<{cls}: {path}>".format(
            cls=self.__class__.__name__,
            path=self.path,
        )

    def __repr__(self):
        return "{cls}('{path}')".format(
            cls=self.__class__.__name__,
            path=self.path,
        )

    @property
    def is_exists(self):
        return os_path.exists(self.path)

    @property
    def is_file(self):
        return os_path.isfile(self.path)

    @property
    def is_dir(self):
        return os_path.isdir(self.path)

    @property
    def is_link(self):
        return os_path.islink(self.path)

    @property
    def dir_path(self):
        return os_path.dirname(self.path)

    @property
    def directory(self):
        return PathInfo(self.dir_path)

    @property
    def children(self):
        if not self.is_dir:
            return []
        return [
            PathInfo(os_path.join(self.path, i))
            for i in os.listdir(self.path)
        ]

    @property
    def full_name(self):
        dir, name = os_path.split(self.path)
        return name

    @property
    def name(self):
        name, ext = os_path.splitext(self.full_name)
        return name

    @property
    def extension(self):
        name, ext = os_path.splitext(self.full_name)
        return ext

    @property
    def stat(self):
        return os.stat(self.path)

    @property
    def length(self):
        return self.stat.st_size

    @property
    def last_access_timestamp(self):
        return self.stat.st_atime

    @property
    def last_access_time(self):
        return datetime.fromtimestamp(self.last_access_timestamp)

    @property
    def last_write_timestamp(self):
        return self.stat.st_mtime

    @property
    def last_write_time(self):
        return datetime.fromtimestamp(self.last_write_timestamp)

    @property
    def creation_timestamp(self):
        return self.stat.st_ctime

    @property
    def creation_time(self):
        return datetime.fromtimestamp(self.creation_timestamp)



#!/usr/bin/env python
# encoding: utf-8

import re

from ycyc.base.adapter import main_entry
from ycyc.base.shelltools import ShellCommands


@main_entry
def main():
    paths = re.findall(
        r"^(?:\s+(?:modified|new file):\s*)(.*?)(?:\s*)$",
        ShellCommands.git.check_output("status"),
        re.M
    )
    if not paths:
        return -1

    return sum(map(
        lambda x: ShellCommands.dos2unix.check_call(str(x)),
        set(paths),
    ))



#!/usr/bin/env python
# encoding: utf-8


class TagMaps(dict):
    DefaultKey = ""

    def __missing__(self, key):
        if isinstance(key, int):
            return self.keys()[key]
        if self.DefaultKey in self:
            return self.get(self.DefaultKey)
        raise KeyError(key)

    def register(self, key, force=False):
        """
        Register obj as a decorator

        :param key: obj key
        :param force: force update
        """
        key = str(key)
        if not force and key in self:
            raise KeyError("%s was existed" % key)

        def update(obj):
            self.update(key, obj)
            return obj

        return update

    def update(self, key_or_dict, obj=NotImplemented):
        if obj is not NotImplemented:
            key_or_dict = {str(key_or_dict): obj}

        return super(TagMaps, self).update(key_or_dict)



#!/usr/bin/env python
# encoding: utf-8

import re


class SimHash(object):
    @classmethod
    def hash_of(cls, tokens, bits):
        """
        Return simhash for n bits by tokens.

        :param tokens: token with weight:[(weight, token), ...]
        :param bits: bits of simhash
        """
        # init the vector
        v_range = tuple(range(bits))
        v_result = [0 for i in v_range]

        # iterate every item of tokens
        for weight, word in tokens:
            # calc the simple hash of word
            word_hash = hash(word)
            # mask to check each bit of word_hash
            bmask = 1
            for i in v_range:
                # when there is 1 on this bit then add
                # the weight of this word to the v_result
                if word_hash & bmask:
                    v_result[i] += weight
                else:
                    v_result[i] -= weight
                # left shift the bmask to check next bit
                bmask <<= 1

        # make fingerprint from v_result
        return tuple(
            1 if i > 0 else 0
            for i in v_result
        )

    def __init__(self, tokens, bits):
        self.hash_result = self.hash_of(tokens, bits)
        self.tokens = tokens
        self.bits = bits

    def raw(self):
        return "".join(self.hash_result)

    def hex(self):
        hex_step = 4
        hex_numbers = "0123456789ABCDEF"
        additional_bits = hex_step - self.bits % hex_step
        if additional_bits < hex_step:
            hash_result = [0] * additional_bits
        else:
            hash_result = []
        hash_result.extend(self.hash_result)

        str_result = []
        for i in range(0, self.bits, hex_step):
            nums = hash_result[i: i + hex_step]
            str_result.append(hex_numbers[reduce(
                lambda n, i: n << 1 | i,
                nums, 0
            )])
        return "".join(str_result)


def simhash(words, bits=128, spliter=None):
    """
    Return simple simhash which weight all is 1 of each item.

    :param words: string words
    :param bits: bits of simhash
    :param spliter: spliter to split words
    """
    if spliter is not None:
        words = spliter(words)
    return SimHash(((1, i) for i in words), bits).raw()


class Spliter(object):
    @classmethod
    def by_space(cls, words):
        return words.split()

    @classmethod
    def by_sep(cls, seps):
        def wrapper(words):
            return words.split(seps)
        return wrapper

    @classmethod
    def by_punctuations(cls, words):
        return re.split(r"\W", words)

    @classmethod
    def by_step(cls, step):
        def wrapper(words):
            word_len = len(words)
            start = 0
            end = step
            while start < word_len:
                yield words[start:end]
                start = end
                end += step

        return wrapper



#!/usr/bin/env python
# encoding: utf-8

"""
funcutils provided some useful functions.
"""

import re
import operator
import inspect


def template_render(template, model):
    """
    A simple template render.
    Example:
    >>> model = {
    ...     "name": "lyc",
    ...     "foo": {
    ...       "bar": "hello"
    ...     }
    ...   }
    >>> template = "{{foo.bar}} {{ name }}"
    >>> template_render(template, model)

    :param template: template string
    :param model: dict model
    """
    parts = []
    for part in re.split(r"{{([\.\w]+?)}}", template):
        parts.append(str(
            part if len(parts) & 1 == 0 else
            reduce(lambda m, k: m[k], part.split("."), model)))
    return "".join(parts)


def encode(s, encoding="utf-8", errors="strict"):
    """
    Auto encode unicode string to bytes.
    :param s: string
    :param encoding: bytes encoding(default:utf-8)
    :param errors: different error handling scheme
    :return: bytes
    """
    if isinstance(s, bytes):
        return s

    return s.encode(encoding, errors)


def decode(s, encoding="utf-8", errors="strict"):
    """
    Auto decode bytes to unicode string.
    :param s: string
    :param encoding: bytes encoding(default:utf-8)
    :param errors: different error handling scheme
    :return: unicode
    """
    if isinstance(s, unicode):
        return s

    return s.decode(encoding, errors)


def bytescat(s1, s2, encoding="utf-8"):
    """
    Auto encode and return s1 + s2
    :param s1: string
    :param s2: string
    :param encoding: bytes encoding(default:utf-8)
    :return: bytes
    """
    return encode(s1, encoding) + encode(s2, encoding)


def strcat(s1, s2, encoding1="utf-8", encoding2=None):
    """
    Auto decode and return s1 + s2
    :param s1: string
    :param s2: string
    :param encoding1: bytes encoding of s1(default:utf-8)
    :param encoding2: bytes encoding of s2(default:same as encoding1)
    :return: bytes
    """
    encoding2 = encoding2 or encoding1
    return decode(s1, encoding1) + decode(s2, encoding2)


def drop_prefix(s, pattern):
    """
    Remove prefix pattern of a string.

    :param s: string
    :param pattern: string pattern
    """
    if s.startswith(pattern):
        return s[len(pattern):]
    return s


def drop_postfix(s, pattern):
    """
    Remove postfix pattern of a string.

    :param s: string
    :param pattern: string pattern
    """
    if s.endswith(pattern):
        return s[:-len(pattern)]
    return s


def split_and_strip(val_str, sep=","):
    """
    Simple split val_str by sep and drop strip the space chars for each items.

    :param val_str: string
    :param sep: split separator
    """
    return [
        i for i in (
            i.strip() for i in val_str.split(sep)
        )
        if i
    ]


def left_part_of(txt, sub_txt, n=1):
    """
    Return the left part before the nth of sub_txt appeared in txt.

    :param txt: text
    :param sub_txt: separate text
    :param n: the nth of sub_txt(default:1)
    """
    parts = txt.split(sub_txt)
    return sub_txt.join(parts[:n])


def right_part_of(txt, sub_txt, n=-1):
    """
    Return the right part after the nth of sub_txt appeared in txt.

    :param txt: text
    :param sub_txt: separate text
    :param n: the nth of sub_txt(default:-1)
    """
    parts = txt.split(sub_txt)
    return sub_txt.join(parts[n:])


class TxtDistance(object):
    @classmethod
    def edit_distance(cls, s1, s2):
        """
        Simple algorithm to calculate the distance between two str

        :param s1: string 1
        :param s2: string 2
        :return: distance number
        """
        len1 = len(s1)
        len2 = len(s2)

        if (len1 <= 0) or (len2 <= 0):
            result = len1 or len2

        row_n = len1 + 1
        col_n = len2 + 1

        f_table = map(lambda x: [0] * col_n, range(row_n))
        for i in range(row_n):
            f_table[i][0] = i
        for i in range(col_n):
            f_table[0][i] = i

        for i in range(1, row_n):
            for j in range(1, col_n):
                if s1[i - 1] == s2[j - 1]:
                    f_table[i][j] = f_table[i - 1][j - 1]
                else:
                    f_table[i][j] = min([
                        f_table[i - 1][j] + 1,
                        f_table[i][j - 1] + 1,
                        f_table[i - 1][j - 1] + 1
                    ])
        return f_table[row_n - 1][col_n - 1]

    @classmethod
    def edit_distance2(cls, str1, str2):
        """
        Simple algorithm to calculate the distance between two str

        :param s1: string 1
        :param s2: string 2
        :return: distance number
        """
        called_result = {}

        def edit_distance(s1, s2):
            result = called_result.get((s1, s2))
            if result:
                return result

            len1 = len(s1)
            len2 = len(s2)
            ss1 = s1[1:]
            ss2 = s2[1:]
            if (len1 <= 0) or (len2 <= 0):
                result = len1 or len2
            elif s1[0] == s2[0]:
                result = edit_distance(ss1, ss2)
            else:
                result = 1 + min(
                    edit_distance(ss1, s2),
                    edit_distance(s1, ss2),
                    edit_distance(ss1, ss2),
                )
            called_result[(s1, s2)] = result
            return result

        return edit_distance(str1, str2)

    @classmethod
    def hamming_distance(cls, s1, s2):
        """
        Reutrn hamming distance between s1 and s2.
        This function assert that len(s1) == len(s2).

        :param s1: string 1
        :param s2: string 2
        :return: distance number
        """
        distance = 0
        for i, j in zip(s1, s2):
            if i != j:
                distance += 1
        return distance


def look_like(target, candidates):
    """
    choice one of string in candidates that target is looks like

    :param target: string
    :param candidates: string list
    :return: item in candidates
    """
    results = [
        (TxtDistance.edit_distance(target, i), i)
        for i in candidates
    ]
    result = min(*results, key=lambda x: x[0])
    return result[1]


def reversed_txt(txt):
    """
    Reverse a txt
    :param txt: string
    :return: reversed string
    """
    return txt[::-1]


def sep_join(sep, sequence, begin="", end=""):
    """
    Separator join each elements of sequence,
    if begin/end is True, insert the sep after
    or before the txt.you can also given some
    txt as head/tail of txt.

    :param sep: separator string
    :param sequence: string sequence
    :param begin: head of txt
    :param end: tail of txt
    :return: reversed string
    """
    txt = sep.join(sequence)
    if begin is True:
        txt = "%s%s" % (sep, txt)
    else:
        txt = "%s%s" % (begin, txt)

    if end is True:
        txt = "%s%s" % (txt, sep)
    else:
        txt = "%s%s" % (txt, end)
    return txt



#!/usr/bin/env python
# encoding: utf-8

import ast

from ycyc.base.contextutils import timeout
from ycyc.base.iterutils import dict_merge

NameParseError = type("NameParseError", (NameError,), {})


class SafeCalc(ast.NodeTransformer):
    """
    Python expression safe calculator
    """
    globals = {
        "locals": None, "globals": None, "__name__": None,
        "__file__": None, "__builtins__": None, "True": True,
        "False": False,
    }

    def __init__(self, locals, timeout=0, interval=None):
        self.locals = locals
        self.timeout = timeout
        self.interval = interval

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name):
            name = node.func.id
            if name not in self.locals and name not in self.globals:
                raise NameParseError("%s not found" % name)
        self.generic_visit(node)
        return node

    def visit_Attribute(self, node):
        name = node.attr
        if name.startswith("__"):
            raise NameParseError("%s not allow" % name)
        self.generic_visit(node)
        return node

    def __call__(self, expr):
        ast_node = ast.parse(expr, mode="eval")
        ast_node = self.visit(ast_node)
        ast_node = ast.fix_missing_locations(ast_node)
        code = compile(ast_node, "<string>", mode="eval")
        with timeout(self.timeout, self.interval):
            return eval(code, self.globals, self.locals)


def safecalc(expr, locals=None, **vars):
    """
    A quick function to calculate Python expression

    :param expr: Python expression
    :param locals: local env
    :param vars: variables
    :return: value
    """
    calc = SafeCalc(dict_merge((vars, locals or {})))
    return calc(expr)



#!/usr/bin/env python
# encoding: utf-8
# source: http://code.activestate.com/recipes/474088-tail-call-optimization-decorator/

import sys
from functools import wraps


class TailRecurseException(Exception):
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs


def tail_call_optimized(g):
    """
    This function decorates a function with tail call
    optimization. It does this by throwing an exception
    if it is it's own grandparent, and catching such
    exceptions to fake the tail call optimization.

    This function fails if the decorated
    function recurses in a non-tail context.
    """
    @wraps(g)
    def func(*args, **kwargs):
        f = sys._getframe()
        if f.f_back and f.f_back.f_back \
                and f.f_back.f_back.f_code == f.f_code:
            raise TailRecurseException(args, kwargs)
        else:
            while True:
                try:
                    return g(*args, **kwargs)
                except TailRecurseException as e:
                    args = e.args
                    kwargs = e.kwargs
    return func



#!/usr/bin/env python
# encoding: utf-8

import functools
import logging

import requests

from ycyc.base.decoratorutils import cachedproperty
from ycyc.base.iterutils import dict_merge

logger = logging.getLogger(__name__)


class Request(requests.Request):
    def __init__(
            self, url, callback, headers=None, params=None, data=None,
            json=None, method="GET", **kwg):
        super(Request, self).__init__(
            url=url, headers=headers, params=params, data=data, json=json,
            method=method, **kwg
        )
        self.callback = callback


class Response(object):
    def __init__(self, raw_request, response):
        self.raw_response = response
        self.raw_request = raw_request

    def __getattr__(self, name):
        return getattr(self.raw_response, name)

    @cachedproperty
    def html(self):
        return self.raw_response.text

    @cachedproperty
    def selector(self):
        import pyquery
        pq = pyquery.PyQuery(self.html)
        return pq

    def make_links_absolute(self):
        self.selector.make_links_absolute(self.raw_response.url)


class Spirder(object):
    def __init__(self, target=None, opener=None, worker=None, headers=()):
        self.worker = worker or self.worker_factory()
        self.opener = opener or requests.Session()
        self.target = target
        self.default_headers = dict(headers)

    def on_request(self, request):
        logger.debug("Spirder on_request, url: %s", request.url)
        self.worker.apply_async(
            self.opener.send,
            args=(request.prepare(),),
            callback=functools.partial(
                self.on_response,
                request=request,
                callback=request.callback,
            ),
        )

    def on_response(self, response, request, callback):
        logger.debug(
            "Spirder on_response, url: %s, status: %s, reason: %s",
            response.url, response.status_code, response.reason,
        )
        real_response = Response(request, response)
        self.add_tasks(callback(real_response))

    def add_tasks(self, requests):
        for request in requests or ():
            logger.debug("Spirder add a new task: %s", request.url)
            request.headers = dict_merge((request.headers, self.default_headers))
            self.worker.apply_async(
                self.on_request,
                args=(request,),
            )

    def start(self):
        logger.info("Spirder start to run")
        requests = self.run()
        self.add_tasks(requests)

    def join(self):
        logger.info("Spirder start to join")
        if hasattr(self.worker, "close"):
            # for multiprocessing.pool
            self.worker.close()
        self.worker.join()

    def run(self):
        if not callable(self.target):
            raise NotImplementedError
        return self.target()

    @classmethod
    def worker_factory(cls):
        from multiprocessing.pool import ThreadPool
        return ThreadPool()


def redirect_to(url, callback, **kwg):
    def redirect(response):
        logger.info("redirect to %s", url)
        yield Request(url=url, callback=callback, **kwg)
    return redirect


def save_to(path, data, mode="ab"):
    with open(path, mode) as fp:
        fp.write(data)


def flow_return():
    raise StopIteration



#!/usr/bin/env python
# encoding: utf-8

from ycyc.base.contextutils import catch
from ycyc.base.typeutils import SimpleExceptions

EventsExceptions = SimpleExceptions()
ListenerDuplicatedError = EventsExceptions.ListenerDuplicatedError
ListenerNoExistedError = EventsExceptions.ListenerNoExistedError
EventNoExistedError = EventsExceptions.EventNoExistedError


class Event(object):
    def __init__(self):
        super(Event, self).__init__()
        self.listeners = []

    def notify(self, *args, **kwg):
        """
        Notify all the listeners that event is happended.
        """
        listeners = tuple(self.listeners)
        for listener in listeners:
            with catch():
                listener(*args, **kwg)

    def register(self, callback):
        """
        Register a callback to this event
        """
        if callback in self.listeners:
            raise ListenerDuplicatedError
        self.listeners.append(callback)

    def unregister(self, callback):
        """
        Unregister a callback from this event
        """
        with catch(ValueError, ListenerNoExistedError):
            self.listeners.remove(callback)


class EventsCenter(dict):
    """
    Event center to manage some events as a dict
    """
    def __init__(self, event_factory=None):
        super(EventsCenter, self).__init__()
        self.event_factory = event_factory or Event

    def __missing__(self, name):
        event = self[name] = self.event_factory()
        return event



#!/usr/bin/env python
# encoding: utf-8

import argparse
import re
import six
import sys
from os import path as os_path
import os
from textwrap import dedent
import logging
from subprocess import CalledProcessError

from ycyc.base.adapter import main_entry
from ycyc.base.resources import Regex
from ycyc.base.filetools import cd
from ycyc.base.shelltools import ShellCommands
from ycyc.base.typeutils import constants
from ycyc.base.logutils import console_only_config

console_only_config()
logger = logging.getLogger(__name__)
ErrorNo = constants(
    UserErr=-1,
    EnvErr=-2,
    SysErr=-3,
    UnknownErr=-4,
)


class Validater(object):
    EmailRex = re.compile("^%s$" % Regex.email_addr())

    @classmethod
    def is_avaliable_email(cls, email):
        return cls.EmailRex.search(email) is not None

    @classmethod
    def is_avaliable_repo(cls, path):
        if not path:
            return False
        if not os_path.isdir(path):
            return False
        if not os_path.isdir(os_path.join(path, ".git")):
            return False
        return True


@main_entry
def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("repo", help="local git repo path")
    arg_parser.add_argument("email", help="email addr in history to filter")
    arg_parser.add_argument("author", help="new author name")
    arg_parser.add_argument("-m", "--new_email", help="new email addr")

    args = arg_parser.parse_args()
    if not Validater.is_avaliable_email(args.email):
        logger.error("[*] %s is not a avaliable email address", args.email)
        return ErrorNo.UserErr
    if not Validater.is_avaliable_repo(args.repo):
        logger.error("[*] %s is not a avaliable repository path", args.repo)
        return ErrorNo.UserErr
    new_email = args.new_email or args.email
    if not Validater.is_avaliable_email(new_email):
        logger.error("[*] %s is not a avaliable email address", new_email)
        return ErrorNo.UserErr

    with cd(args.repo):
        logger.info("[-] change directory to: %s" % os.getcwd())
        try:
            return ShellCommands.git.check_call(
                "filter-branch", "-f", "--env-filter",
                dedent("""
                    if [ "$GIT_COMMITTER_EMAIL" = "{old_email}" ]; then
                        export GIT_COMMITTER_NAME="{new_author}"
                        export GIT_COMMITTER_EMAIL="{new_email}"
                    fi
                    if [ "$GIT_AUTHOR_EMAIL" = "{old_email}" ]; then
                        export GIT_AUTHOR_NAME="{new_author}"
                        export GIT_AUTHOR_EMAIL="{new_email}"
                    fi""".format(
                    old_email=args.email,
                    new_author=args.author,
                    new_email=new_email,
                )),
                "--tag-name-filter", "cat", "--",
                "--branches", "--tags"
            )
        except CalledProcessError as err:
            logger.error("[x] returned code of git: %s", err.returncode)
            return ErrorNo.EnvErr
        except Exception as err:
            logger.error("[?] %s", str(err))
            return ErrorNo.UnknownErr



#!/usr/bin/env python
# encoding: utf-8

import contextlib
import logging.config
import logging
import sys
import thread
import inspect

logger = logging.getLogger(__name__)


class LoggerInfo(object):
    def __init__(self, frame=None):
        if frame is None:
            frames = sys._current_frames()
            frame = frames[thread.get_ident()]
            self.frame = frame.f_back
        else:
            self.frame = frame

    @property
    def line_no(self):
        return self.frame.f_lineno

    @property
    def code_name(self):
        return self.frame.f_code.co_name

    @property
    def is_top_frame(self):
        if self.frame.f_back is None:
            return True
        return False

    @property
    def func_args(self):
        args = inspect.getargvalues(self.frame)
        code_locals = args.locals
        result = {
            i: code_locals[i]
            for i in args.args
        }
        if args.varargs:
            result[args.varargs] = code_locals[args.varargs]
        if args.keywords:
            result[args.keywords] = code_locals[args.keywords]
        return result


def quick_config(log_file="application.log"):
    """
    a quick config for logging

    :param log_file: the file for file handler
    """
    config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'verbose': {
                'format': '%(asctime)s %(levelname)s %(name)s %(message)s',
                'datefmt': '[%d/%b/%Y %H:%M:%S]',
            },
            'simple': {
                'format': '%(asctime)s %(levelname)s %(name)s:%(lineno)s %(message)s',
                'datefmt': '[%d/%b/%Y %H:%M:%S]',
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
            'file': {
                'level': 'INFO',
                'class': 'logging.FileHandler',
                'filename': log_file,
                'formatter': 'verbose'
            },
        },
        'loggers': {
            '': {
                'handlers': ['console', 'file'],
                'level': 'DEBUG',
            },
        },
    }
    logging.config.dictConfig(config)


def console_only_config(level='DEBUG', format='%(message)s'):
    config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'simple': {
                'format': format,
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
        },
        'loggers': {
            '': {
                'handlers': ['console'],
                'level': 'DEBUG',
            },
        },
    }
    logging.config.dictConfig(config)


@contextlib.contextmanager
def log_disable(level=logging.CRITICAL):
    """
    A context manager to disable all logging calls of
    severity 'level' and below in block
    """
    cur_frames = sys._current_frames()
    cur_frame = cur_frames[thread.get_ident()]
    pre_frame = cur_frame.f_back

    logger.info(
        "with logger disabled from %s to %s in %s[%s], id: %s",
        logging.root.manager.disable, level,
        pre_frame.f_code.co_name, pre_frame.f_lineno,
        id(cur_frame),
    )

    level, logging.root.manager.disable = logging.root.manager.disable, level
    try:
        yield
    finally:
        level, logging.root.manager.disable = logging.root.manager.disable, level

        logger.info("logger reseted, id: %s", id(cur_frame))


def log_with_label(log_method, label):
    """
    >>> logger_waring = log_with_label(logger.waring, "waring: %s")
    >>> try:
    ...     may_raise_exception()
    ... except Exception as err:
    ...     logger_waring(err)

    :param log_method: method of a logger to pack
    :param label: label pass to log_method as the first argument
    :return: logger method packed method
    """
    import functools

    @functools.wraps(log_method)
    def logger_method(*args, **kwargs):
        return log_method(label, *args, **kwargs)

    return logger_method


class LogFunctionCall(object):
    def __init__(self, func=None, args=None, kwargs=None, returned=None):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.returned = returned

    def __str__(self):
        s_returned = (
            "" if self.returned is None
            else "-> %s" % str(self.returned)
        )

        args = map(str, self.args or ())
        args.extend(
            "{k}={v}".format(k=str(k), v=str(v))
            for k, v in self.kwargs.items() or ()
        )

        return "{f}({args}){returned}".format(
            f=self.func.__name__ if self.func else "",
            args=", ".join(args),
            returned="" if self.returned is None
            else "-> %s" % str(self.returned)
        )



#!/usr/bin/env python
# encoding: utf-8

import types
import functools
import importlib
import sys


class LazyEnv(object):
    """
    LazyEnv() is a environment manager which accept a lambda and
    get the result lazily.Use it as module global value manager
    to avoid recursive dependency of modules.
    Example:
        # module_a
        import module_b
        from lazyutils import LazyEnv
        GlobalEnv = LazyEnv()
        GlobalEnv.val = "a"
        GlobalEnv.txt = lambda: "i'm not " + module_b.GlobalEnv.val

        # module_b
        import module_a
        from lazyutils import LazyEnv
        GlobalEnv = LazyEnv()
        GlobalEnv.val = "b"
        GlobalEnv.txt = lambda: "i'm not " + module_a.GlobalEnv.val

    Anyway,it is not a suggested solution.
    You should use 'from xxx import xx' statement carefully.

    You can set a value immediately to LazyEnv(),there are two ways
    when you want to set a lambda:
        GlobalEnv.val = lambda: lambda: 1
        GlobalEnv["val"] = lambda: 1
    But LazyEnv() is not a dict like object.
    """
    def __setattr__(self, attr, val):
        if hasattr(self, attr):
            delattr(self, attr)
        super(LazyEnv, self).__setattr__("_lazy_%s" % attr, val)

    def __getattr__(self, attr):
        lazy_attr = "_lazy_%s" % attr
        if hasattr(self, lazy_attr):
            val = getattr(self, lazy_attr)
            if isinstance(val, types.LambdaType):
                val = val()
            super(LazyEnv, self).__setattr__(attr, val)
            delattr(self, lazy_attr)
            return val
        raise AttributeError("%s not found" % attr)

    def __setitem__(self, key, val):
        setattr(self, key, lambda: val)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError("%s not found" % key)


def lazy_init(func):
    """
    Decorator for __getattr__ method only, cached the real __getattr__
    method returned value and init the attribute.
    """
    assert func.func_name == "__getattr__"

    @functools.wraps(func)
    def lazy_get(self, name):
        value = func(self, name)
        setattr(self, name, value)
        return value

    return lazy_get


class FakeModule(object):
    @classmethod
    def reload(cls, module):
        """
        Reload the fake module
        """
        real_module = sys.modules[module.__name__]
        reload(real_module)
        return module


def lazy_import(module_name):
    """
    Lazy to import a module
    """
    def getattribute(self, attr):
        module = importlib.import_module(module_name)
        self.__dict__ = module.__dict__
        return getattr(module, attr)

    return type(
        FakeModule.__name__,
        (FakeModule,),
        {"__getattribute__": getattribute}
    )()
